#!/usr/bin/env python3

#
# Author: Bioinformatics group, Takara Bio USA, Inc.
#
# ....
#

# -----------------------------------------------------------------------------
# region: Script Info

SCRIPT_NAME = "select_barcode"
MAIN_DESC = """
	(Demo Version) Select barcodes after demultiplexing and do downsampling if required.
	"""

# Do not put indent, because it affects for output.
SUB_DESC = """
description:
  Demo version.
"""
# endregion
# -----------------------------------------------------------------------------

# ---------- import modules ---------- #

import shutil
import sys
from glob import glob
from multiprocessing import Manager, Pool
from os import mkdir, path, unlink
from os.path import isdir
from time import sleep
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from common.cogent_error import CogentError
from common.logger import Logger
from common.util.demux import DemuxSummary, load_summary_file
from common.util.util import run_system_cmd

# -----------------------------------------------------------------------------
# region: Function


# Check all input requirements
def check_requirements(user_args) -> bool:
    res = True

    # Check if input path (directory) exists
    if not isdir(user_args.input_dir):
        Logger.error_both(f"Input directory does not exist. {user_args.input_dir}")
        res = False

    return res


def find_original_count_csv(target_dir: str) -> Union[str, None]:
    if isdir(target_dir):  # Assume target dir was generated by demux command
        count_file = glob(
            f"{target_dir}/*_counts_all_original.csv"
        )  # Try to find backup up file.

        if len(count_file) > 1:
            raise CogentError(f"Too many counts_all_original.csv in {target_dir}")
        elif len(count_file) == 0:
            return None
        else:
            return count_file[0]  # Return file
    else:
        raise CogentError(f"Unexpected user input type", "find_original_count_csv")


def find_usable_count_csv(target_dir: str) -> Union[str, None]:
    if isdir(target_dir):  # Assume target dir was generated by demux command
        count_file = glob(
            f"{target_dir}/*_counts_all.csv"
        )  # Try to find current count csv.

        if len(count_file) > 1:
            raise CogentError(f"Too many counts_all.csv in {target_dir}")  # Return file
        elif len(count_file) == 0:
            return None
        else:
            return count_file[0]
    else:
        raise CogentError(f"Unexpected user input type", "find_usable_count_csv")


def find_count_csv_auto(target_dir: str) -> str:
    if isdir(target_dir):  # Assume target dir was generated by demux command
        count_file = find_original_count_csv(
            target_dir
        )  # Try to find backup up file first.

        if count_file is None:
            count_file = find_usable_count_csv(
                target_dir
            )  # Try to find current count csv.

            if count_file is None:
                raise CogentError(
                    f"Cannot find neither counts_all.csv nor counts_all_original.csv in {target_dir}"
                )

        return count_file  # Return file
    else:
        raise CogentError(f"Unexpected user input type", "find_count_csv_auto")


def show_stats(data: np.ndarray):
    # Show stats
    pd.options.display.precision = 0
    print()
    print(pd.DataFrame(pd.Series(data).describe()).transpose())
    print()


def get_input(
    message: str,
    available: Union[List[str], None] = None,
    value_range: Union[Tuple[int, int], None] = None,
) -> str:
    input_value = ""

    if available is not None:
        while input_value not in available:
            print(f"{message}: ", end="")
            input_value = input().strip()

    elif value_range is not None:
        temp_input = -1
        while temp_input < value_range[0] or temp_input > value_range[1]:
            print(f"{message} ({value_range[0]} to {value_range[1]}): ", end="")

            try:
                n = float(input().strip())
                if n.is_integer():
                    temp_input = int(n)
                else:
                    print(f"Please input correct integer.")
            except ValueError:
                print(f"Please input correct integer.")

        input_value = str(temp_input)
    else:
        print(f"{message}: ", end="")
        input_value = input().strip()

    return input_value


def select_barcodes(
    original: Dict[str, DemuxSummary], select_type: str, number: int
) -> List[DemuxSummary]:
    new_barcodes = None

    if select_type == "1":
        non_samples = [
            summary for summary in original.values() if summary.name == "Non_sample"
        ]  # Get only non samples first
        samples = [
            summary for summary in original.values() if summary.name != "Non_sample"
        ]  # Get samples only
        samples = sorted(
            samples, key=lambda x: x.reads_num, reverse=True
        )  # Sort list by read numbers in reverse order
        new_barcodes = samples[0:number]  # Select top X
        new_barcodes = new_barcodes + non_samples  # Join non_samples
    elif select_type == "2":
        new_barcodes = [
            summary
            for summary in original.values()
            if summary.reads_num >= number or summary.name == "Non_sample"
        ]  # Retain Non_sample always
    else:
        raise CogentError(f"Unknown selecting type: {select_type}")

    return new_barcodes


def collect_fastq_files(
    barcodes: List[DemuxSummary], target_dir: str
) -> Dict[str, List[str]]:
    ret = {}
    process_count = 0
    total_count = len(barcodes)
    for data in barcodes:
        fastq_files = glob(
            f"{target_dir}/*_{data.key}_{data.name}_R*.fastq.gz"
        )  # glob R1 and R2 for the barcode
        if len(fastq_files) == 1 or len(fastq_files) == 2:
            ret[data.key] = fastq_files
        else:
            raise CogentError(
                f"Cannot collect FASTQ files for downsampling: {data.key}"
            )

        process_count += 1
        draw_progressbar(process_count, total_count)

    print()

    return ret


def do_downsampling(
    pid: int,
    key: str,
    files: List[Tuple[str, str]],
    reads_num: int,
    res_dict: Dict[str, List[int]],
) -> bool:
    try:
        # Run Seqtk
        for in_out in files:
            res = run_system_cmd(
                " ".join(["seqtk", "sample", in_out[0], str(reads_num)]),
                stdout_file=in_out[1],
            )  # Tentative. Need to refere CogentAP_tools
            if not res:
                return False

        # Count FASTQ files
        count_list = []
        for in_out in files:
            # Count output FASTQ
            line_count = 0
            with open(in_out[1], "r") as f:
                for _ in f:
                    line_count += 1

            if line_count % 4 != 0:
                return False

            count_list.append(int(line_count / 4))

        # Check FASTQ file
        count_set = set(count_list)
        if len(count_set) != 1:
            return False

        res_dict[key] = count_list

        # Compress output FASTQ file
        for in_out in files:
            res = run_system_cmd(" ".join(["pigz", "-p", "1", in_out[1]]))
            if not res:
                return False

    except KeyboardInterrupt:
        pass  # Do nothing

    return True


def process_downsample(
    in_files: Dict[str, List[str]], out_dir: str, reads_num: int
) -> bool:
    pool = Pool(processes=8)  # Use literal for now

    try:
        Logger.info_both("Start downsampling ...")

        manager = Manager()  # Create manager instance to receive input
        result_dict = manager.dict()  # Create dict for output from each process

        # Launch subprocesses
        res = {}
        pid = 1
        complete_list = []
        for key, files in in_files.items():
            out_files = [
                (f, f"{out_dir}/{path.splitext(path.basename(f))[0]}") for f in files
            ]

            res[pid] = pool.apply_async(
                do_downsampling, args=(pid, key, out_files, reads_num, result_dict)
            )
            pid += 1

        # Wait for complete
        while len(complete_list) != len(res.keys()):
            # Check process status
            for p, r in res.items():
                if p in complete_list:
                    continue

                if r.ready():
                    complete_list.append(p)

            # Redraw progress bar
            draw_progressbar(len(complete_list), len(res.keys()))

            sleep(1)

        pool.close()
        pool.join()
        print()

        Logger.info_both("Downsampling completed.")

        # Write out temporary read numbers
        with open(f"{out_dir}/read_count.tmp", "w") as f:
            for key, count in result_dict.items():
                f.write(",".join([key, str(count[0])]) + "\n")  # type: ignore

        return all(
            [r.get() for r in res.values()]
        )  # Check and retrun all sub process returned True

    except KeyboardInterrupt:
        Logger.warning_both(
            "Received Ctrl+C to terminate process. Try to terminate subprocesses."
        )
        pool.terminate()
        pool.join()
        return False


def draw_progressbar(current_count: int, max_count: int):
    BAR_LENGTH = 80
    progress = current_count / max_count

    eq_num = int(BAR_LENGTH * progress)
    space_num = BAR_LENGTH - eq_num
    arrow_str = ""
    if space_num > 0:
        space_num = space_num - 1
        arrow_str = ">"

    print(
        f'\r{str(int(progress*100)).rjust(3, " ")}% [{"=" * eq_num}{arrow_str}{" " * space_num}] ({current_count}/{max_count})',
        end="\r",
    )


# endregion: Function
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# region: Main Function
def run(user_args, first_log):
    # # ---------- setup | logger ---------- #
    Logger.initialize(SCRIPT_NAME, f"{SCRIPT_NAME}.log")
    Logger.info_logger(first_log)
    Logger.info_logger("Command name: " + SCRIPT_NAME)
    Logger.info_logger("Original call : " + " ".join(sys.argv))

    # ---------- setup | check requirement ---------- #
    if not check_requirements(user_args):
        return False

    # ---------- Find count.csv ---------- #
    input_dir = user_args.input_dir.rstrip("/")
    count_csv = find_count_csv_auto(input_dir)
    Logger.info_both(f"Found count.csv: {count_csv}")
    
    output_dir = user_args.results_dir.rstrip("/")

    # ---------- Load count.csv ---------- #
    original_barcodes = load_summary_file(count_csv)
    Logger.info_both(
        f"Loaded {len(original_barcodes.keys()) - 3} barcodes."
    )  # Ignore 3 records (Short, Unselected, Undetermined) to count
    original_reads = np.array(
        [
            summary.reads_num
            for summary in original_barcodes.values()
            if summary.name != "Non_sample"
        ]
    )

    # ---------- Show stats ---------- #
    Logger.info_both("Show original stats.")
    show_stats(original_reads)

    complete_select = False
    retained_barcodes = None
    retained_reads = None
    while not complete_select:
        # ---------- Receive input ---------- #
        Logger.info_both(
            "Please input selecting type. (1: Select N barcodes from top, 2: Select barcodes with N reads or more.)"
        )
        select_type = get_input("Selecting type (1 or 2)", available=["1", "2"])
        Logger.info_both(f"Selecting type is {select_type}.")

        select_number = -1
        if select_type == "1":
            Logger.info_both("Please input number of barcodes to be retained.")
            select_number = int(
                get_input("Number of barcodes", value_range=(1, original_reads.size))
            )
        elif select_type == "2":
            Logger.info_both("Please input number of minimum reads to be retained.")
            select_number = int(
                get_input(
                    "Number of minimum reads", value_range=(1, original_reads.max())
                )
            )
        else:
            raise CogentError(f"Unknown selecting type: {select_type}")

        Logger.info_both(f"Input number to select is {select_number}.")

        # ---------- Calculate result of selecting ---------- #
        retained_barcodes = select_barcodes(
            original_barcodes, select_type, select_number
        )
        retained_reads = np.array(
            [
                summary.reads_num
                for summary in retained_barcodes
                if summary.name != "Non_sample"
            ]
        )

        # ---------- Show stats after selecting ---------- #
        Logger.info_both("Show stats of retained barcodes.")
        show_stats(retained_reads)

        # ---------- Confirm ---------- #
        answer = get_input(
            "Do you want to proceed steps with these barcodes? (yes/no)",
            available=["Yes", "No", "yes", "no"],
        )

        if answer.lower() == "yes":
            complete_select = True
        else:
            Logger.info_both("Restart barcode selection.")

    if retained_barcodes is None:
        raise CogentError(f"Unknown error. Retained barcodes information is empty.")

    # Check file backup
    current_count_csv = find_usable_count_csv(input_dir)
    original_count_csv = find_original_count_csv(input_dir)
    if original_count_csv is None:  # When original does not exist
        # Create backup
        shutil.copy(current_count_csv, f"{input_dir}/{path.splitext(path.basename(current_count_csv))[0]}_original.csv")  # type: ignore

        # Get backup file again
        original_count_csv = find_original_count_csv(input_dir)
        if original_count_csv is None:
            raise CogentError(f"Cannot create backup of count.csv")

        Logger.info_both(
            f"Backup count.csv: {current_count_csv} -> {original_count_csv}"
        )
    else:
        Logger.info_both(f"Backup file already exists. No need to backup again.")

    # Select barcodes in new list from original barcode list to keep same order as input
    target_keys = [summary.key for summary in retained_barcodes]
    output_list = [
        summary for summary in original_barcodes.values() if summary.key in target_keys
    ]

    # ---------- Confirm downsampling ---------- #
    final_barcodes = None
    answer = get_input(
        "Do you want to do downsampling? (yes/no)", available=["Yes", "No", "yes", "no"]
    )
    if answer.lower() == "yes":
        # With downsampling
        keep_reads = int(
            get_input(
                "Please input number of reads to keep as maximum",
                value_range=(1, original_reads.max()),
            )
        )

        Logger.info_both(f"Collecting FASTQ files ...")
        in_fastq_files = collect_fastq_files(
            [data for data in output_list if data.name != "Non_sample"], input_dir
        )

        # Create output directory
        out_dir = output_dir
        if isdir(out_dir):  # When directory is exist
            Logger.info_both(f"Output directory for downsampling exists: {out_dir}")
            answer = get_input(
                f"Are you sure to overwrite {out_dir}? (yes/no)",
                available=["Yes", "No", "yes", "no"],
            )
            if answer.lower() == "yes":
                Logger.info_both(f"Delete directory: {out_dir}")
                shutil.rmtree(out_dir)
            else:
                Logger.info_both(f"Terminate process.")
                return True

        Logger.info_both(f"Create directory: {out_dir}")
        mkdir(out_dir)

        # Downsampling
        res = process_downsample(in_fastq_files, out_dir, keep_reads)
        if not res:
            Logger.error_both(f"Downsampling failed.")
            return False

        # Load downsample result from temp file
        downsample_result = {}
        with open(f"{out_dir}/read_count.tmp", "r") as in_f:
            for line in in_f:
                values = line.rstrip().split(",")
                downsample_result[values[0]] = values[1]

        # Write out new count.csv
        with open(f"{out_dir}/demux_ds_counts_all.csv", "w") as out_f:
            for data in output_list:
                if data.name == "Non_sample":
                    out_f.write(
                        ",".join([data.key, data.name, str(data.reads_num)]) + "\n"
                    )
                else:
                    out_f.write(
                        ",".join([data.key, data.name, downsample_result[data.key]])
                        + "\n"
                    )

        final_barcodes = load_summary_file(find_usable_count_csv(out_dir))  # type: ignore
        unlink(f"{out_dir}/read_count.tmp")
    else:
        # Without downsampling
        answer = get_input(
            "Are you sure to overwrite current count.csv? (yes/no)",
            available=["Yes", "No", "yes", "no"],
        )
        if answer.lower() == "yes":
            with open(current_count_csv, "w") as out_f:  # type: ignore
                for data in output_list:
                    out_f.write(
                        ",".join([data.key, data.name, str(data.reads_num)]) + "\n"
                    )
        else:
            Logger.info_both(f"Terminate process.")
            return True

        final_barcodes = load_summary_file(find_usable_count_csv(input_dir))  # type: ignore

    # ---------- Show final stats ---------- #
    final_reads = np.array(
        [
            data.reads_num
            for data in final_barcodes.values()
            if data.name != "Non_sample"
        ]
    )
    Logger.info_both("Selecting done. Show final stats.")
    show_stats(final_reads)

    return True


# endregion
# -----------------------------------------------------------------------------
